/*******************************************************************

    YAPP(tm) protocol for Packet Radio binary file transfer

	(c) 1986, Jeff Jacobsen - WA7MBL

	Revision 1.1   06/23/86

	----------------------------------------------------------

	IW3FQG's Yapp Implementation

	This source include resume on download

    (c) Copyright 1990/1993 Marco Savegnago IW3FQG

*******************************************************************/

#define DEBUGTYAPP

/* pklib include */

#include <xferprot/tyapp.h>
#include <dos/ttimeout.h>

#define __MISC_INLINE   // usa routine inline
#include <dos/tmisc.h>

/* stdlib include */

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <io.h> 	// per access

#if 0
#define DEBUGTYAPP
#	include "mstream.h"
#endif

yappTStr TYapp::strRcvReady = { ACK, SOH, NIL };	// RR
yappTStr TYapp::strRcvFile  = { ACK, STX, NIL };	// RF
yappTStr TYapp::strAckEOF   = { ACK, ETX, NIL };	// AF     [Yapp 1.1]
yappTStr TYapp::strAckEOT   = { ACK, EOT, NIL };	// AT	  [Yapp 1.1]
yappTStr TYapp::strSendInit = { ENQ, SOH, NIL };	// SN

yappTStr TYapp::strSendEOF  = { ETX, SOH, NIL };	// EF
yappTStr TYapp::strSendEOT  = { EOT, SOH, NIL };	// ET
yappTStr TYapp::strCanACK   = { ACK, ENQ, NIL };	// CA
yappTStr TYapp::strNotReady = { NAK, NIL, NIL };	// NR
yappTStr TYapp::strCancel	= { CAN, NIL, NIL };	// CN


#ifdef __YAPPUSEDDVT

void TYapp::Dispatch( yapp_states state )
{
	ConFP theDDVTfoo;           // final location for DDVT foo

	union
    {
		FP    fp;               // since the RTL returns a basic
		ConFP memberFP;         // function pointer address, we
	} temp;                     // use a union to transform to
								// member function pointer.

	FP fp = *_DDVTdispatchNULL( GetVptr( this ) , state );

	if ( fp )
	{
		temp.fp = fp;
		theDDVTfoo = temp.memberFP;
	}

	(this->*theDDVTfoo)();
}

#elif defined (__YAPPUSESWITCH)

/*  Using a big switch ... dispatcher */

void TYapp::Dispatch( yapp_states state )
{
	switch (state)
	{
		case	SN:
			sendInit();
			break;

		case 	SR:
			sendInitRetry();
			break;

		case	SH:
			sendHdr();
			break;

		case 	SD:
			sendData();
			break;

		case 	SE:
			sendEOF();
			break;

		case	ST:
			sendEOT();
			break;

		case 	RC:
			receive();
			break;

		case	RH:
			recvHdr();
			break;

		case	RD:
			recvData();
			break;

		case	AB:
			cancel();
			break;

		case	CW:
			canWait();
			break;

		case	CR:
			canRecd();
			break;

		default:
			break;

	}
}

#endif

Boolean TYapp::waitReady()
{

	/*******************************************************************

    If there is already something, simply return ....

    *******************************************************************/

	if (theStream->isAny())
    				return False;


    /*******************************************************************

    Else activate activate a timeout

    *******************************************************************/

    TTimeout timeout;

	if ( state==SN || state==SR )
#ifdef  DEBUGTYAPP
				timeout.setTimer(90);		// 5 secondi debug...
#else
				timeout.setTimer(364);		// 20 sec
#endif
	else
				timeout.setTimer(2184);		// or 120 sec

	// While timeout is not expired and the port is connected

	while(!timeout && theStream->isConnected())
		if (theStream->isAny())
				return False;

	ptype = TM;

	return True;
}

void TYapp::getPacket()
{
	ptype = UK; // Unknow

	if (waitReady())
				return;

	char ch = theStream->get();

	switch(ch)
	{
		case ACK:
			if (waitReady())
						return;

			ch = theStream->get();

			switch (ch)
			{
				case SOH:
					ptype=RR;
					break;
				case STX:
					ptype=RF;
					break;
				case ETX:
					ptype=AF;
					break;
				case EOT:
					ptype=AT;
					break;
				case ENQ:
					ptype=CA;
					break;
				default:
					break;
			}

			return;

		case ENQ:
			if (waitReady())
						return;

			ch = theStream->get();

			switch (ch)
			{
				case SOH:
					ptype=SI;   	// Get a SendInit from other station
					break;

				case STX:
					ptype=RI;		// Get a ReceiveInit from other station
					break;

				default:
					ptype=UU;		// Get an unknow type from other station
					break;
			}

			return;

		case SOH:
			ptype=HD;
			getPacketString();
			break;

		case STX:
			ptype=DT;
			getPacketString();
			break;

		case ETX:
			if (waitReady())
						return;

			ch = theStream->get();

			if (ch == SOH)
					ptype = EF;
			break;

		case EOT:
			if (waitReady())
						return;

			ch = theStream->get();

			if (ch == SOH)
					ptype = ET;

			break;

		case NAK:
			ptype=NR;
			getPacketString();
			break;

		case CAN:
			ptype=CN;
			getPacketString();
			break;

		case DLE:
			ptype=TX;
			getPacketString();
			break;
		default:
			break;
	}

	return;
}

void TYapp::getPacketString()
{
	// is there something?

	if (waitReady())
				return;

	// Yes. get it!

	frameLen=(theStream->get() & 0xFF);

	if (ptype == DT && !frameLen)
			frameLen = maxYappFrame;

	if (!frameLen)
			return;

    for (int i = 0; i < frameLen; i++)
    {
		if (waitReady())
					return;

		*(frameBuffer+i)=theStream->get();
    }

}

void TYapp::sendInit()
{
/*******************************************************************

	Send SendInit String

*******************************************************************/

	*theStream << strSendInit;

	theStream->flush();

/*******************************************************************

	Wait for responce

*******************************************************************/

	getPacket();

/*******************************************************************

	Bring appropriate actions

*******************************************************************/

	switch (ptype)
	{
		case TM:
			state=SR;
			break;

		case RI:
			state=SN;
			break;

		case RR:
			state=SH;
			break;

		case RF:
			state=SD;
			break;

		case CN:
			state=CR;
			break;

		case NR:
			state=SA;
			break;

		case TX:
			dispPacket();
			break;

		default:
			if (state==AB)
			{
#ifdef DEBUGTYAPP
		mout << "Abort by originator" << endl;
#endif
				break;
			}
			state=AB;
#ifdef DEBUGTYAPP
		mout << "Protocol Error" << endl;
#endif
			break;
	}
}

void TYapp::sendInitRetry()
{
	if(++sendInitCount > maxRetry)
	{
		state=AB; // Abort

#ifdef DEBUGTYAPP
		mout << "Timeout" << endl;
#endif

		return;
	}

	sendInit();
}

void TYapp::sendHdr()
{
/*******************************************************************
	Reset XFER counter
*******************************************************************/

	xferCount=0L;

	if(	theStream->put(frameBuffer, 2 + makeHeader()) == False)
    {
#ifdef DEBUGTYAPP
		mout << "Get False from TStream::put\n" << endl;
#endif
	}

	theStream->flush();

/*******************************************************************

	Wait responce

*******************************************************************/

	getPacket();

	switch(ptype)
	{
		case RF:
			state=SD;
			break;
		case CN:
			state=CR;
			break;
		case NR:
            if(yappResume && (*frameBuffer=='R'))
            {
				long resumeFrom = atol((char *)frameBuffer+2);

                xferCount=resumeFrom;

				if ((xferCount=resumeFrom)==fstreamLenght(theFileStream))
				{
					state=SE;
					break;
				}

                theFileStream->seekg(xferCount=resumeFrom);

                if(theFileStream->fail())
                {
                	state=AB;
                    break;
                }

#ifdef DEBUGTYAPP
				mout << "File resumed from " << xferCount << endl;
#endif
				// Else send file from the position of resource

				state=SD;
			}
			else
			{
#ifdef DEBUGTYAPP
				mout << "Abort by destination" << endl;
#endif
				state=AB;
			}

			break;

		case TX:
			dispPacket();
			break;

		default:
			if (state==AB)
			{
#ifdef DEBUGTYAPP
			mout << "Abort by originator" << endl;
#endif
				break;
			}

			state=AB;

			if (ptype==TM)
			{
#ifdef DEBUGTYAPP
		mout << "Timeout" << endl;
#endif
			;
            }
			else
            {
#ifdef DEBUGTYAPP
		mout << "Protocol Error" << endl;
#endif
			;
            }
			break;
	}
}

void TYapp::sendData()
{
#ifdef DEBUGTYAPP
		mout << "YAPP: SENDDATA isAny?\n" << endl;
#endif
	if (theStream->isAny())
	{
		getPacket();

		if ((ptype == CN))
		{
			state = CR;
			return;
		}
		else if ((ptype == TX))
			dispPacket();
		else
		{
			state = AB;
			return;
		}
   }

   yappTFrame yappFrame;

   word cnt=maxYappFrame;

   if(!read(yappFrame.frame, cnt))
   {
		// lastError=ReadingFile;
		state=AB;
        return;
   }

#ifdef DEBUGTYAPP
	if(!theFileStream)
		mout << "Error reading " << endl;
#endif

   if (cnt)
   {
	  yappFrame.count = (cnt == maxYappFrame) ? 0 : cnt;

	  if(theStream->put((byte *) &yappFrame, cnt + 2) == False)
      {
#ifdef DEBUGTYAPP
		mout << "Get False from TStream::put\n" << endl;
#endif
      	state = AB;
		return;
      }
	  xferCount += (long) cnt;
   }

#ifdef DEBUGTYAPP
		mout << "Byte transferred: " << xferCount << endl;
#endif

	// If the frame is < of maxYappFrame then eot

    if (cnt != maxYappFrame)
					state = SE;
}

void TYapp::sendEOF()
{
#ifdef DEBUGTYAPP
		mout << "File sent" << endl;
#endif
	*theStream << strSendEOF;

	theStream->flush();

	getPacket();

	switch(ptype)
	{
		case AF:
            xferOK=True;
			state = ST;
			break;

		case TX:
			dispPacket();
			break;

		default:

			state=AB;
			if (ptype==TM)
            {
#ifdef DEBUGTYAPP
				mout << "Timeout" << endl;
#endif
				;
            }
			else
            {
#ifdef DEBUGTYAPP
				mout << "Protocol Error" << endl;
#endif
				;
            }
			break;
	}
}

void TYapp::sendEOT()
{
	*theStream << strSendEOT;

	theStream->flush();

	getPacket();

	if(ptype==TX)
		dispPacket();
	else
		state=SA;
}

void TYapp::receive()
{
	getPacket();

	switch (ptype)
	{
		case SI:
			*theStream << strRcvReady;
			theStream->flush();

			state=RH;
			break;

		case CN:
			state=CR;
			break;

		case TX:
			dispPacket();
			break;

		default:
			if (state==AB)
			{
#ifdef DEBUGTYAPP
				mout << "Abort by destination" << endl;
#endif
				break;
			}
			state=AB;
			if(ptype==TM)
            {
#ifdef DEBUGTYAPP
				mout << "Timeout" << endl;
#endif
            ;
			}
			else
            {
#ifdef DEBUGTYAPP
				mout << "Protocol Error" << endl;
#endif
				;
            }
			break;
	}
}


void TYapp::recvHdr()
{

#ifdef DEBUGTYAPP
		mout << "Now receive header" << endl;
#endif

	getPacket();

	switch(ptype)
	{
		case HD:
			xferOK=False;
			xferCount=0L;

#ifdef DEBUGTYAPP
        	splitHeader();

			mout << "Header : " << yappHeaderFileName << " " << yappHeaderFileLenght << endl;
#endif

			*theStream << strRcvFile;

			theStream->flush();

			state=RD;
			break;

		case SI:
			state=RH;
			break;

		case CN:
			state=CR;
			break;

		case ET:
			*theStream << strAckEOT;
			theStream->flush();
			state=SA;
			break;

		case TX:
			dispPacket();
			break;

		default:
			if (state==AB)
			{
#ifdef DEBUGTYAPP
				mout << "Abort by destination" << endl;
#endif
				break;
			}
			state=AB;

			if (ptype==TM)
            {
#ifdef DEBUGTYAPP
				mout << "Timeout" << endl;
#endif
			;
            }
			else
            {
#ifdef DEBUGTYAPP
				mout << "Protocol error" << endl;
#endif
			;
            }
	}
}

void TYapp::recvData()
{

#ifdef DEBUGTYAPP
		mout << "Receiving file" << endl;
#endif

    getPacket();

	switch(ptype)
	{
		case DT:

            if(!write(frameBuffer, frameLen))
            {
				// lastError=WritingFile;
				state=AB;
        		return;
			}

            xferCount += (long) frameLen;

#ifdef DEBUGTYAPP
			mout << "Byte trasferred: " << xferCount << endl;
#endif
 			state=RD;
			break;

		case EF:
            xferOK=True;

			*theStream << strAckEOF;

			theStream->flush();

#ifdef DEBUGTYAPP
			mout << "Receive complete" << endl;
#endif

			state=RH;
			break;

		case CN:

#ifdef DEBUGTYAPP
			mout << "Abort from originator" << endl;
#endif

			state=CR;
			break;

		case TX:
			dispPacket();
			break;

		default:
			if (state==AB)
			{

#ifdef DEBUGTYAPP
				mout << "Abort from destination" << endl;
#endif
				break;
			}
			state=AB;

			if(ptype==TM)
            {
#ifdef DEBUGTYAPP
				mout << "Timeout" << endl;
#endif
			;
            }
			else
            {
#ifdef DEBUGTYAPP
				mout << "Protocol error" << endl;
#endif
			;
            }
			break;
	}
}

void TYapp::cancel()
{
	if(theStream->put((byte*) strCancel,2)==False)
    {
#ifdef DEBUGTYAPP
		mout << "YAPP CANCEL: Get False from TStream::put\n" << endl;
#endif
    }
  	state=CW;
}

void TYapp::canWait()
{
	getPacket();

	switch(ptype)
	{
		case CA:
			state=SA;
			break;
		case CN:
			*theStream << strCanACK;

			theStream->flush();

			break;
		case TM:
		case UK:
			state=SA;
			break;
		case TX:
			dispPacket();
			break;
		default:
			break;
	}
}

void TYapp::canRecd()
{
#ifdef DEBUGTYAPP
	mout << "Abort from originator" << endl;
#endif
	*theStream << strCanACK;

	theStream->flush();

  	state=SA;
}


Boolean TYapp::splitHeader()
{
	strcpy(yappHeaderFileName, (char *)frameBuffer);

    char fileLen[16];

	strcpy(fileLen, (char *)(frameBuffer+strlen(yappHeaderFileName)+2));

	yappHeaderFileLenght=atol(fileLen);

	return True;
}

/*******************************************************************

	Make header for send a file:

	A standard Yapp header is as follow:

	<SOH> <len> (Filename) <NUL> (File Size in ASCII) <NUL>

	If extended header is selected then the following field are appended
	to the standard header.

	(Filedatetime) <NUL> (Fileattribute) <NUL>

	Il campo Filedatetime e' espesso mediante una stringa hex
	contenete un numero a 32 bit che al suo interno e' ripartito
	nei seguenti campi:

		 31 30 29 28 27 26 25 24 23 22 21 20 19 18 17 16
		[<---- anno-1980 --->|<---mese-->|<---giorno--->]

		 15 14 13 12 11 10  9  8  7  6  5  4  3  2  1  0
		[<----ore----->|<-----minuti----->|<-secondi/2->]

	che corrispondono rispettivamente al contenuto dei registri
	CX e DX al ritorno della funzione DOS 57h.
		(disponibili nella funzione getftime() in TurboC )

	Il campo Fileattribute e' espresso con una stringa hex a 8 bit
	il cui significato e' il seguente:

		Bit             Attributi
		---             ---------
		0               File di sola lettura
		1               File nascosto
		2               File di sistema
		5               File di archivio

	che corrispondono al contenuto del reg. CX nella funzione 43h.
		(disponibili nella funzione setftime() in TurboC )

*******************************************************************/

byte TYapp::makeHeader()
{
	memset(frameBuffer, '\0', sizeof(yappFrameBuffer));

	stripPath(theCurrentFilename, yappHeaderFileName);

	strcpy((char *)frameBuffer + 2, yappHeaderFileName);

	fileSize=yappHeaderFileLenght=fstreamLenght(theFileStream);

	char fileLen[16];

	ltoa(yappHeaderFileLenght,fileLen,10);

	strcpy((char *)(frameBuffer + 2 + strlen(yappHeaderFileName) + 1), fileLen);

	byte frameLen=byte(strlen(yappHeaderFileName)+strlen(fileLen) + 2);

	if(extendedHeader)
	{
		getftime((theFileStream->rdbuf())->fd(),&yappHeaderFiletime);

		char daytime[20];

		char access[10];

		sprintf(daytime,"%x",&yappHeaderFiletime);

		sprintf(access,"%x",(char)_chmod(theCurrentFilename,0));

		strcpy((char *)(frameBuffer + frameLen + 1 ), daytime);

		frameLen += strlen(daytime) + 1;

		strcpy((char *)(frameBuffer + frameLen), access);

		frameLen += strlen(access) + 1;
	}

	*frameBuffer=SOH;

	*(frameBuffer+1)=frameLen;

	return frameLen;
}


TYapp::TYapp(TStream *port , fstream *afstream, xferType type, char *filename, Boolean resume) :
	TBinaryXferProtcol(port, afstream, type)
{
	theCurrentFilename=strdup(filename);
	sendInitCount=0;

	state = (type == Receive) ? RC : SN;

	yappResume=resume;

	extendedHeader = False;
}


TYapp::~TYapp()
{
	free(theCurrentFilename);
/*
	if(!xferOK)
		theFileStream->close();

*/
}

Boolean TYapp::xfer()
{

	// Save the old port mode

	TStreamMode old = theStream->setMode(moUnPolled|moBinary);

	// do the transfer

	do Dispatch(state);

	while (state != SA);


    // reset the old port mode

    theStream->setMode(old);


    // return to the caller with the trasfert result

	return xferOK;
}

#if defined (__YAPPUSEFPTR)
conFP TYapp::stateDispatchTable[] =
{
	TYapp::sendInit,
	TYapp::sendInitRetry,
	TYapp::sendHdr,
	TYapp::sendData,
	TYapp::sendEOF,
	TYapp::sendEOT,
	TYapp::receive,
	TYapp::recvHdr,
	TYapp::recvData,
	TYapp::cancel,
	TYapp::canWait,
	TYapp::canRecd
};

#endif

